# Generated by Django 5.2.6 on 2025-10-03 01:36

import django.core.validators
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('django_ledger', '0026_stagedtransactionmodel_customer_model_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='accountmodel',
            name='role',
            field=models.CharField(
                choices=[
                    (
                        'Assets',
                        [
                            ('asset_ca_cash', 'Current Asset'),
                            ('asset_ca_mkt_sec', 'Marketable Securities'),
                            ('asset_ca_recv', 'Receivables'),
                            ('asset_ca_inv', 'Inventory'),
                            ('asset_ca_uncoll', 'Uncollectibles'),
                            ('asset_ca_prepaid', 'Prepaid'),
                            ('asset_ca_other', 'Other Liquid Assets'),
                            ('asset_lti_notes', 'Notes Receivable'),
                            ('asset_lti_land', 'Land'),
                            ('asset_lti_sec', 'Securities'),
                            ('asset_ppe_build', 'Buildings'),
                            ('asset_ppe_build_accum_depr', 'Buildings - Accum. Depreciation'),
                            ('asset_ppe_plant', 'Plant'),
                            ('asset_ppe_plant_depr', 'Plant - Accum. Depreciation'),
                            ('asset_ppe_equip', 'Equipment'),
                            ('asset_ppe_equip_accum_depr', 'Equipment - Accum. Depreciation'),
                            ('asset_ia', 'Intangible Assets'),
                            ('asset_ia_accum_amort', 'Intangible Assets - Accum. Amortization'),
                            ('asset_adjustment', 'Other Assets'),
                        ],
                    ),
                    (
                        'Liabilities',
                        [
                            ('lia_cl_acc_payable', 'Accounts Payable'),
                            ('lia_cl_credit_line', 'Credit Line'),
                            ('lia_cl_wages_payable', 'Wages Payable'),
                            ('lia_cl_int_payable', 'Interest Payable'),
                            ('lia_cl_taxes_payable', 'Taxes Payable'),
                            ('lia_cl_st_notes_payable', 'Short Term Notes Payable'),
                            ('lia_cl_ltd_mat', 'Current Maturities of Long Tern Debt'),
                            ('lia_cl_def_rev', 'Deferred Revenue'),
                            ('lia_cl_other', 'Other Liabilities'),
                            ('lia_ltl_notes', 'Long Term Notes Payable'),
                            ('lia_ltl_bonds', 'Bonds Payable'),
                            ('lia_ltl_mortgage', 'Mortgage Payable'),
                        ],
                    ),
                    (
                        'Equity',
                        [
                            ('eq_capital', 'Capital'),
                            ('eq_stock_common', 'Common Stock'),
                            ('eq_stock_preferred', 'Preferred Stock'),
                            ('eq_adjustment', 'Other Equity Adjustments'),
                            ('eq_dividends', 'Dividends & Distributions to Shareholders'),
                            ('in_operational', 'Operational Income'),
                            ('in_passive', 'Investing/Passive Income'),
                            ('in_interest', 'Interest Income'),
                            ('in_gain_loss', 'Capital Gain/Loss Income'),
                            ('in_other', 'Other Income'),
                            ('cogs_regular', 'Cost of Goods Sold'),
                            ('ex_regular', 'Regular Expense'),
                            ('ex_interest_st', 'Interest Expense - Short Term Debt'),
                            ('ex_interest', 'Interest Expense - Long Term Debt'),
                            ('ex_taxes', 'Tax Expense'),
                            ('ex_capital', 'Capital Expense'),
                            ('ex_depreciation', 'Depreciation Expense'),
                            ('ex_amortization', 'Amortization Expense'),
                            ('ex_other', 'Other Expense'),
                        ],
                    ),
                    (
                        'Root',
                        [
                            ('root_coa', 'CoA Root Account'),
                            ('root_assets', 'Assets Root Account'),
                            ('root_liabilities', 'Liabilities Root Account'),
                            ('root_capital', 'Capital Root Account'),
                            ('root_income', 'Income Root Account'),
                            ('root_cogs', 'COGS Root Account'),
                            ('root_expenses', 'Expenses Root Account'),
                        ],
                    ),
                ],
                max_length=30,
                verbose_name='Account Role',
            ),
        ),
        migrations.AlterField(
            model_name='receiptmodel',
            name='amount',
            field=models.DecimalField(
                decimal_places=2,
                help_text='Amount of the receipt.',
                max_digits=20,
                validators=[django.core.validators.MinValueValidator(limit_value=0)],
                verbose_name='Receipt Amount',
            ),
        ),
        migrations.AlterField(
            model_name='receiptmodel',
            name='receipt_account',
            field=models.ForeignKey(
                blank=True,
                help_text='The income or expense account where this transaction will be reflected',
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to='django_ledger.accountmodel',
                verbose_name='PnL Account',
            ),
        ),
        migrations.AlterField(
            model_name='receiptmodel',
            name='receipt_type',
            field=models.CharField(
                choices=[
                    ('sales', 'Sales Receipt'),
                    ('customer_refund', 'Sales Refund'),
                    ('expense', 'Expense Receipt'),
                    ('expense_refund', 'Expense Refund'),
                    ('transfer', 'Transfer Receipt'),
                    ('debt_paydown', 'Debt Paydown Receipt'),
                ],
                max_length=15,
                verbose_name='Receipt Type',
            ),
        ),
        migrations.AlterField(
            model_name='stagedtransactionmodel',
            name='receipt_type',
            field=models.CharField(
                blank=True,
                choices=[
                    ('sales', 'Sales Receipt'),
                    ('customer_refund', 'Sales Refund'),
                    ('expense', 'Expense Receipt'),
                    ('expense_refund', 'Expense Refund'),
                    ('transfer', 'Transfer Receipt'),
                    ('debt_paydown', 'Debt Paydown Receipt'),
                ],
                help_text='The receipt type of the transaction.',
                max_length=20,
                null=True,
                verbose_name='Receipt Type',
            ),
        ),
    ]



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Receipt models and utilities.

This module provides models and helpers for recording monetary receipts within an
Entity's ledger. It supports receipts related to customers (sales and refunds),
vendors (expenses and refunds), and transfers between accounts. In addition to
Django ORM models, the module exposes queryset/manager helpers for filtering
receipts, utilities for generating sequential receipt numbers, and migration of
staged bank transactions into posted journal entries.

Notes
-----
- All public functions and methods are documented in NumPy docstring style.
- Unless otherwise noted, monetary amounts are Decimal-compatible and are
  validated server-side.
"""

from datetime import datetime
from decimal import Decimal
from typing import Literal, Optional
from uuid import UUID, uuid4

from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.core.validators import MinValueValidator
from django.db import IntegrityError, models, transaction
from django.db.models import F, Manager, Q, QuerySet
from django.db.models.signals import pre_save
from django.urls import reverse
from django.utils.timezone import localdate
from django.utils.translation import gettext_lazy as _

from django_ledger.io.io_core import IOMixIn
from django_ledger.models import (
    AccountModel,
    CreateUpdateMixIn,
    CustomerModel,
    EntityModel,
    EntityStateModel,
    EntityUnitModel,
    MarkdownNotesMixIn,
    VendorModel,
)
from django_ledger.settings import (
    DJANGO_LEDGER_DOCUMENT_NUMBER_PADDING,
    DJANGO_LEDGER_RECEIPT_NUMBER_PREFIX,
)


class ReceiptModelValidationError(ValidationError):
    """Domain-specific validation error for receipt operations.

    This exception is raised when invalid states or inputs are detected while
    creating, configuring, validating, or mutating receipt instances.
    """

    pass


class ReceiptModelQuerySet(QuerySet):
    def for_user(self, user_model) -> 'ReceiptModelQuerySet':
        """Filter receipts visible to a given user.

        Parameters
        ----------
        user_model : django.contrib.auth.models.User
            The user for whom to filter receipts. The user must be either the
            entity admin or listed among the entity managers.

        Returns
        -------
        ReceiptModelQuerySet
            QuerySet containing receipts associated with entities the user can
            manage.
        """
        return self.filter(
            Q(ledger_model__entity__admin=user_model) | Q(ledger_model__entity__managers__in=[user_model])
        )

    def for_dates(self, from_date, to_date) -> 'ReceiptModelQuerySet':
        """Filter receipts within an inclusive date range.

        Parameters
        ----------
        from_date : date
            Start date (inclusive).
        to_date : date
            End date (inclusive).

        Returns
        -------
        ReceiptModelQuerySet
            QuerySet containing receipts between the specified dates.
        """
        return self.filter(receipt_date__gte=from_date, receipt_date__lte=to_date)

    def for_vendor(self, vendor_model: VendorModel | str | UUID) -> 'ReceiptModelQuerySet':
        """Filter receipts tied to a specific vendor.

        Parameters
        ----------
        vendor_model : VendorModel or str or UUID
            The vendor instance, vendor number (str), or vendor UUID to filter
            by. Only expense-related receipts are returned.

        Returns
        -------
        ReceiptModelQuerySet
            QuerySet containing receipts associated with the given vendor.

        Raises
        ------
        ReceiptModelValidationError
            If the provided value is not a supported type.
        """
        if isinstance(vendor_model, str):
            return self.filter(
                vendor_model__vendor_number__iexact=vendor_model,
                customer_model__isnull=True,
            )
        elif isinstance(vendor_model, VendorModel):
            return self.filter(
                vendor_model=vendor_model,
                customer_model__isnull=True,
            )
        elif isinstance(vendor_model, UUID):
            return self.filter(
                vendor_model_id=vendor_model,
                customer_model__isnull=True,
            )
        raise ReceiptModelValidationError(
            'Invalid Vendor Model: {}, must be instance of VendorModel, UUID, str'.format(vendor_model)
        )

    def for_customer(self, customer_model: CustomerModel | str | UUID) -> 'ReceiptModelQuerySet':
        """Filter receipts tied to a specific customer.

        Parameters
        ----------
        customer_model : CustomerModel or str or UUID
            The customer instance, customer number (str), or customer UUID to
            filter by. Only sales-related receipts are returned.

        Returns
        -------
        ReceiptModelQuerySet
            QuerySet containing receipts associated with the given customer.

        Raises
        ------
        ReceiptModelValidationError
            If the provided value is not a supported type.
        """
        if isinstance(customer_model, str):
            return self.filter(
                customer_model__customer_number__iexact=customer_model,
                vendor_model__isnull=True,
            )
        elif isinstance(customer_model, CustomerModel):
            return self.filter(
                customer_model=customer_model,
                vendor_model__isnull=True,
            )
        elif isinstance(customer_model, UUID):
            return self.filter(
                customer_model_id=customer_model,
                vendor_model__isnull=True,
            )
        raise ReceiptModelValidationError(
            'Invalid Customer Model: {}, must be instance of CustomerModel, UUID, str'.format(customer_model)
        )


class ReceiptModelManager(Manager):
    def get_queryset(self):
        """Return the base queryset with common annotations.

        Returns
        -------
        ReceiptModelQuerySet
            A queryset pre-annotated with entity UUID, slug, and last closing
            date, and with ledger_model selected.
        """
        return (
            ReceiptModelQuerySet(self.model, using=self._db)
            .select_related('ledger_model')
            .annotate(
                _entity_uuid=F('ledger_model__entity__uuid'),
                _entity_slug=F('ledger_model__entity__slug'),
                _last_closing_date=F('ledger_model__entity__last_closing_date'),
            )
        )

    def for_entity(self, entity_model: EntityModel | str | UUID) -> ReceiptModelQuerySet:
        """Filter receipts for a specific entity.

        Parameters
        ----------
        entity_model : EntityModel or str or UUID
            The entity instance, slug, or UUID to filter receipts for.

        Returns
        -------
        ReceiptModelQuerySet
            Receipts belonging to the specified entity.

        Raises
        ------
        ReceiptModelValidationError
            If an unsupported type is provided.
        """
        qs = self.get_queryset()
        if isinstance(entity_model, str):
            qs = qs.filter(ledger_model__entity__slug__exact=entity_model)
        elif isinstance(entity_model, UUID):
            qs = qs.filter(ledger_model__entity_id=entity_model)
        elif isinstance(entity_model, EntityModel):
            qs = qs.filter(ledger_model__entity=entity_model)
        else:
            raise ReceiptModelValidationError(
                f'Must pass either EntityModel, string or UUID, not {type(entity_model)}.'
            )
        return qs


class ReceiptModelAbstract(CreateUpdateMixIn, MarkdownNotesMixIn, IOMixIn):
    """Abstract model representing a monetary receipt.

    This class encapsulates the common fields and behaviors for receipts across
    different types: sales, sales refund, expense, expense refund, and transfer.
    It includes helpers to generate sequential receipt numbers, validate receipt
    configuration, and migrate staged transactions to journal entries.
    """

    SALES_RECEIPT = 'sales'
    SALES_REFUND = 'customer_refund'
    EXPENSE_RECEIPT = 'expense'
    EXPENSE_REFUND = 'expense_refund'
    TRANSFER_RECEIPT = 'transfer'
    DEBT_PAYMENT = 'debt_paydown'

    RECEIPT_TYPES = [
        (SALES_RECEIPT, 'Sales Receipt'),
        (SALES_REFUND, 'Sales Refund'),
        (EXPENSE_RECEIPT, 'Expense Receipt'),
        (EXPENSE_REFUND, 'Expense Refund'),
        (TRANSFER_RECEIPT, 'Transfer Receipt'),
        (DEBT_PAYMENT, 'Debt Paydown Receipt'),
    ]
    RECEIPT_TYPES_MAP = dict(RECEIPT_TYPES)

    uuid = models.UUIDField(default=uuid4, editable=False, primary_key=True)
    receipt_number = models.CharField(_('Receipt Number'), max_length=255)
    receipt_date = models.DateField(_('Receipt Date'))
    receipt_type = models.CharField(choices=RECEIPT_TYPES, max_length=15, verbose_name=_('Receipt Type'))

    ledger_model = models.ForeignKey(
        'django_ledger.LedgerModel',
        on_delete=models.PROTECT,
        verbose_name=_('Ledger Model'),
        editable=False,
    )

    unit_model = models.ForeignKey(
        'django_ledger.EntityUnitModel',
        on_delete=models.PROTECT,
        verbose_name=_('Unit Model'),
        help_text=_('Helps segregate receipts and transactions into different classes or departments.'),
        null=True,
        blank=True,
    )

    customer_model = models.ForeignKey(
        'django_ledger.CustomerModel',
        on_delete=models.PROTECT,
        verbose_name=_('Customer Model'),
        null=True,
        blank=True,
    )
    vendor_model = models.ForeignKey(
        'django_ledger.VendorModel',
        on_delete=models.PROTECT,
        verbose_name=_('Vendor Model'),
        null=True,
        blank=True,
    )

    charge_account = models.ForeignKey(
        'django_ledger.AccountModel',
        on_delete=models.PROTECT,
        verbose_name=_('Charge Account'),
        help_text=_('The financial account (cash or credit) where this transaction was made.'),
        related_name='charge_receiptmodel_set',
    )

    receipt_account = models.ForeignKey(
        'django_ledger.AccountModel',
        on_delete=models.PROTECT,
        verbose_name=_('PnL Account'),
        help_text=_('The income or expense account where this transaction will be reflected'),
        null=True,
        blank=True,
    )

    amount = models.DecimalField(
        decimal_places=2,
        max_digits=20,
        verbose_name=_('Receipt Amount'),
        help_text=_('Amount of the receipt.'),
        validators=[MinValueValidator(limit_value=0)],
    )

    staged_transaction_model = models.OneToOneField(
        'django_ledger.StagedTransactionModel',
        on_delete=models.RESTRICT,
        null=True,
        blank=True,
        verbose_name=_('Staged Transaction Model'),
        help_text=_('The staged transaction associated with the receipt from bank feeds.'),
    )

    objects = ReceiptModelManager.from_queryset(queryset_class=ReceiptModelQuerySet)()

    class Meta:
        abstract = True
        verbose_name = _('Sales/Expense Receipt')
        verbose_name_plural = _('Sales/Expense Receipts')
        indexes = [
            models.Index(fields=['receipt_number']),
            models.Index(fields=['ledger_model']),
            models.Index(fields=['receipt_date']),
            models.Index(fields=['receipt_type']),
            models.Index(fields=['customer_model']),
            models.Index(fields=['vendor_model']),
        ]

    @property
    def entity_slug(self) -> str:
        """Entity slug convenience property.

        Returns
        -------
        str
            The slug of the related entity, using the annotated value when
            present to avoid extra queries.
        """
        try:
            return getattr(self, '_entity_slug')
        except AttributeError:
            pass
        return self.ledger_model.entity_slug

    @property
    def entity_uuid(self) -> UUID:
        """Entity UUID convenience property.

        Returns
        -------
        UUID
            The UUID of the related entity, using the annotated value when
            available to minimize queries.
        """
        try:
            return getattr(self, '_entity_uuid')
        except AttributeError:
            pass
        return self.ledger_model.entity_uuid

    @property
    def ledger_posted(self):
        """Whether the underlying ledger is posted.

        Returns
        -------
        bool
            True if the associated ledger is posted, False otherwise.
        """
        return self.ledger_model.is_posted()

    @property
    def posted(self):
        """Alias for ledger_posted for template convenience.

        Returns
        -------
        bool
            True if the associated ledger is posted, False otherwise.
        """
        return self.ledger_model.is_posted()

    @property
    def last_closing_date(self):
        """Entity last closing date convenience property.

        Returns
        -------
        date
            The entity's last closing date, using the annotated value when
            available to minimize queries.
        """
        try:
            return getattr(self, '_last_closing_date')
        except AttributeError:
            pass
        return self.ledger_model.entity.last_closing_date

    # Actions
    def can_delete(self) -> bool:
        """Check if the receipt can be safely deleted.

        Returns
        -------
        bool
            True if the receipt date is after the entity's last closing date.
        """
        return all(
            [
                self.last_closing_date < self.receipt_date,
            ]
        )

    def delete(self, using=None, keep_parents=False, delete_ledger: bool = True, **kwargs):
        """Delete the receipt and related journal entries if allowed.

        Parameters
        ----------
        using : str, optional
            The database alias to use.
        keep_parents : bool, default False
            Whether to keep parent model records.
        delete_ledger : bool, default True
            Unused flag preserved for backward compatibility.
        **kwargs
            Additional options passed to Django's delete.

        Returns
        -------
        tuple
            A 2-tuple of the number of objects deleted and a dictionary with the
            number of deletions per object type, as returned by Django.

        Raises
        ------
        ReceiptModelValidationError
            If the receipt is within a closed period.
        """
        if not self.can_delete():
            raise ReceiptModelValidationError(
                message=_('Receipt cannot be deleted because it falls within a closed period.'),
            )
        ledger = self.ledger_model
        with transaction.atomic():
            if ledger.is_locked():
                ledger.unlock(commit=True, raise_exception=True)

            ledger.journal_entries.all().delete()

            return super().delete(using=using, keep_parents=keep_parents)

    def is_sales_receipt(self) -> bool:
        """Whether the receipt is sales-related (sales or refund).

        Returns
        -------
        bool
            True if receipt_type is SALES_RECEIPT or SALES_REFUND.
        """
        return any(
            [
                self.receipt_type == self.SALES_RECEIPT,
                self.receipt_type == self.SALES_REFUND,
            ]
        )

    def is_expense_receipt(self) -> bool:
        """Whether the receipt is expense-related (expense or refund).

        Returns
        -------
        bool
            True if receipt_type is EXPENSE_RECEIPT or EXPENSE_REFUND.
        """
        return any(
            [
                self.receipt_type == self.EXPENSE_RECEIPT,
                self.receipt_type == self.EXPENSE_REFUND,
            ]
        )

    def is_transfer_receipt(self) -> bool:
        """Whether the receipt is a transfer between accounts.

        Returns
        -------
        bool
            True if receipt_type is TRANSFER_RECEIPT.
        """
        return self.receipt_type == self.TRANSFER_RECEIPT

    def get_receipt_type_for_amount(self, amount: float | int | Decimal) -> str:
        """Derive a receipt type from context and amount sign.

        Parameters
        ----------
        amount : float or int or Decimal
            The receipt amount. The sign determines whether a refund or normal
            receipt is implied given the counterparty type.

        Returns
        -------
        str
            One of SALES_RECEIPT, SALES_REFUND, EXPENSE_RECEIPT, EXPENSE_REFUND.

        Raises
        ------
        ReceiptModelValidationError
            If both a customer and vendor are set, or neither is set.

        Notes
        -----
        Rules:
        - Customer: amount >= 0 -> SALES_RECEIPT; amount < 0 -> SALES_REFUND
        - Vendor:   amount <= 0 -> EXPENSE_RECEIPT; amount > 0 -> EXPENSE_REFUND
        """
        if self.customer_model_id and self.vendor_model_id:
            raise ReceiptModelValidationError(
                message='Cannot determine receipt type when both customer and vendor are set.'
            )

        if self.customer_model_id:
            return self.SALES_RECEIPT if float(amount) >= 0 else self.SALES_REFUND

        if self.vendor_model_id:
            return self.EXPENSE_REFUND if float(amount) > 0 else self.EXPENSE_RECEIPT

        raise ReceiptModelValidationError(message='Cannot determine receipt type without a customer or vendor.')

    def is_configured(self) -> bool:
        """Whether the receipt has enough data to operate.

        Returns
        -------
        bool
            True if receipt_date, receipt_type, and ledger_model are set.
        """
        return all(
            [
                self.receipt_date is not None,
                self.receipt_type is not None,
                self.ledger_model_id is not None,
            ]
        )

    def can_generate_receipt_number(self) -> bool:
        """Whether a receipt number can be generated for this instance.

        Returns
        -------
        bool
            True if the receipt currently has no receipt_number assigned.
        """
        return all([not self.receipt_number])

    def _get_next_state_model(self, raise_exception: bool = True):
        """Fetch or create the next sequential state for receipt numbering.

        Parameters
        ----------
        raise_exception : bool, default True
            If True, re-raise any IntegrityError encountered during creation.

        Returns
        -------
        EntityStateModel or None
            The state model containing the updated sequence number or None when
            an IntegrityError occurs and raise_exception=False.
        """
        entity_model = self.ledger_model.entity
        fy_key = entity_model.get_fy_for_date(dt=self.receipt_date)

        LOOKUP = {
            'entity_model_id__exact': entity_model.uuid,
            'entity_unit_id__exact': None,
            'fiscal_year': fy_key,
            'key__exact': EntityStateModel.KEY_RECEIPT,
        }

        try:
            state_model_qs = (
                EntityStateModel.objects.filter(**LOOKUP).select_related('entity_model').select_for_update()
            )

            state_model = state_model_qs.get()
            state_model.sequence = F('sequence') + 1
            state_model.save(update_fields=['sequence'])
            state_model.refresh_from_db()
            return state_model
        except ObjectDoesNotExist:
            LOOKUP = {
                'entity_model_id': entity_model.uuid,
                'entity_unit_id': None,
                'fiscal_year': fy_key,
                'key': EntityStateModel.KEY_RECEIPT,
                'sequence': 1,
            }

            state_model = EntityStateModel.objects.create(**LOOKUP)
            return state_model
        except IntegrityError as e:
            if raise_exception:
                raise e

    def generate_receipt_number(self, commit: bool = False) -> str:
        """Generate and optionally persist a sequential receipt number.

        Parameters
        ----------
        commit : bool, default False
            If True, persist the newly generated number to the database.

        Returns
        -------
        str
            The generated or existing receipt number.
        """
        if self.can_generate_receipt_number():
            state_model = None
            while not state_model:
                state_model = self._get_next_state_model(raise_exception=False)

            seq = str(state_model.sequence).zfill(DJANGO_LEDGER_DOCUMENT_NUMBER_PADDING)
            self.receipt_number = f'{DJANGO_LEDGER_RECEIPT_NUMBER_PREFIX}-{state_model.fiscal_year}-{seq}'

            if commit:
                self.save(update_fields=['receipt_number', 'updated'])

        return self.receipt_number

    def configure(
        self,
        entity_model: EntityModel | str | UUID,
        receipt_type: Literal[SALES_RECEIPT, SALES_REFUND, EXPENSE_RECEIPT, EXPENSE_REFUND, TRANSFER_RECEIPT],
        amount: int | float | Decimal,
        unit_model: Optional[EntityUnitModel | str | UUID] = None,
        receipt_date: Optional[datetime | str] = None,
        vendor_model: Optional[VendorModel | str | UUID] = None,
        customer_model: Optional[CustomerModel | str | UUID] = None,
        charge_account: Optional[AccountModel] = None,
        receipt_account: Optional[AccountModel] = None,
        staged_transaction_model=None,
        commit: bool = True,
    ):
        """Configure the receipt with entity, counterpart, and accounting data.

        This method sets related models, validates inputs, creates a posted ledger
        for the receipt, and generates a receipt number.

        Parameters
        ----------
        entity_model : EntityModel or str or UUID
            The entity instance, slug, or UUID that owns the receipt.
        receipt_type : {SALES_RECEIPT, SALES_REFUND, EXPENSE_RECEIPT, EXPENSE_REFUND}
            The type of receipt.
        amount : int or float or Decimal
            Positive monetary amount of the receipt.
        unit_model : EntityUnitModel or str or UUID, optional
            Unit/class/department context.
        receipt_date : datetime or str, optional
            Date to use for the receipt; defaults to local date.
        vendor_model : VendorModel or str or UUID, optional
            The vendor related to the receipt (mutually exclusive with customer).
        customer_model : CustomerModel or str or UUID, optional
            The customer related to the receipt (mutually exclusive with vendor).
        charge_account : AccountModel, optional
            The cash/credit account where the transaction occurred.
        receipt_account : AccountModel, optional
            The income/expense account impacted by the receipt.
        staged_transaction_model : StagedTransactionModel, optional
            Optional staged transaction to link for migration.
        commit : bool, default True
            If True, persist the configured receipt to the database.

        Raises
        ------
        ReceiptModelValidationError
            On invalid combinations or values.
        """
        if not self.is_configured():
            with transaction.atomic():
                if amount < 0:
                    raise ReceiptModelValidationError(message='Receipt amount must be greater than zero')
                if isinstance(entity_model, EntityModel):
                    pass
                elif isinstance(entity_model, UUID):
                    entity_model = EntityModel.objects.get(uuid__exact=entity_model)
                elif isinstance(entity_model, str):
                    entity_model = EntityModel.objects.get(slug__exact=entity_model)

                self.receipt_type = receipt_type
                self.amount = amount
                self.receipt_date = localdate() if not receipt_date else receipt_date
                self.charge_account = charge_account
                self.receipt_account = receipt_account
                self.unit_model = unit_model
                self.staged_transaction_model = staged_transaction_model

                if self.is_transfer_receipt():
                    if any([vendor_model, customer_model]):
                        raise ReceiptModelValidationError(
                            message='Transfer receipts do not require a vendor model or customer model',
                        )
                else:
                    if all([vendor_model, customer_model]):
                        raise ReceiptModelValidationError(
                            message='Must pass VendorModel or CustomerModel, not both.',
                        )

                    if not any([vendor_model, customer_model]):
                        raise ReceiptModelValidationError(
                            message='Must pass VendorModel or CustomerModel.',
                        )

                    # checks if a vendor model has been previously assigned....
                    if all(
                        [
                            vendor_model is not None,
                            self.vendor_model_id is not None,
                        ]
                    ):
                        raise ReceiptModelValidationError(message='Vendor Model already set.')

                    # checks if a customer model has been previously assigned....
                    if all(
                        [
                            customer_model is not None,
                            self.customer_model_id is not None,
                        ]
                    ):
                        raise ReceiptModelValidationError(message='Customer Model already set.')

                # get vendor model...
                if vendor_model:
                    if isinstance(vendor_model, str):
                        vendor_model = VendorModel.objects.for_entity(entity_model=entity_model).get(
                            vendor_number__iexact=vendor_model
                        )
                    elif isinstance(customer_model, UUID):
                        vendor_model = VendorModel.objects.for_entity(entity_model=entity_model).get(
                            uuid__exact=vendor_model
                        )
                    elif isinstance(vendor_model, VendorModel):
                        vendor_model.validate_for_entity(entity_model=entity_model)
                    else:
                        raise ReceiptModelValidationError(
                            message='VendorModel must be either a VendorModel, UUID or Vendor Number.'
                        )
                    self.vendor_model = vendor_model

                # get customer model
                if customer_model:
                    if isinstance(customer_model, str):
                        customer_model = CustomerModel.objects.for_entity(entity_model=customer_model).get(
                            customer_number__iexact=customer_model
                        )
                    elif isinstance(customer_model, UUID):
                        customer_model = CustomerModel.objects.for_entity(entity_model=customer_model).get(
                            uuid__exact=customer_model
                        )
                    elif isinstance(customer_model, CustomerModel):
                        customer_model.validate_for_entity(entity_model=entity_model)
                    else:
                        raise ReceiptModelValidationError(
                            message='Customer Model must be either a CustomerModel, UUID or Customer Number.'
                        )
                    self.customer_model = customer_model

                if unit_model:
                    if isinstance(unit_model, str):
                        unit_model = EntityUnitModel.objects.for_entity(entity_model=entity_model).get(
                            slug__exact=unit_model
                        )
                    elif isinstance(unit_model, UUID):
                        unit_model = EntityUnitModel.objects.for_entity(entity_model=entity_model).get(
                            uuid__exact=unit_model
                        )
                    elif isinstance(unit_model, EntityUnitModel):
                        unit_model.validate_for_entity(entity_model=entity_model)

                self.ledger_model = entity_model.create_ledger(
                    name=entity_model.name,
                    posted=True,
                    commit=False,
                )

                receipt_number = self.generate_receipt_number(commit=False)

                self.ledger_model.name = receipt_number
                self.ledger_model.save()
                self.clean()

                if commit:
                    self.save()

    def can_migrate(self) -> bool:
        """Whether the receipt has enough data to migrate staged transactions.

        Returns
        -------
        bool
            True if a receipt_date is set and either vendor or customer is set.
        """
        return all(
            [
                self.receipt_date is not None,
                self.amount > 0.00,
                self.charge_account_id is not None,
                any(
                    [
                        self.vendor_model_id is not None if not self.is_transfer_receipt() else True,
                        self.customer_model_id is not None if not self.is_transfer_receipt() else True,
                    ]
                ),
            ]
        )

    def migrate_receipt(self, split_amount: bool = False):
        """Post staged transactions into the ledger as journal entries.

        This method commits staged transactions linked to the receipt into the
        associated ledger, marking them as posted journal entries.

        Raises
        ------
        ReceiptModelValidationError
            If the receipt is not configured or lacks the required counterparty
            to migrate (vendor or customer).
        """
        if not self.is_configured():
            raise ReceiptModelValidationError(
                message='Receipt Model must be configured. Call configure() before migrating receipt.',
            )
        if not self.can_migrate():
            raise ReceiptModelValidationError(
                message='Must have VendorModel or CustomerModel, not both.',
            )

        commit_dict = self.staged_transaction_model.commit_dict(split_txs=split_amount)
        ledger_model = self.ledger_model
        staged_to_save = list()
        for je_data in commit_dict:
            _, _ = ledger_model.commit_txs(
                je_timestamp=self.receipt_date,
                je_unit_model=self.unit_model,
                je_posted=True,
                je_desc=f'Receipt Number: {self.receipt_number}',
                je_origin='migrate_receipt',
                je_txs=je_data,
            )
            staged_to_save += [i['staged_tx_model'] for i in je_data]
        staged_to_save = set(staged_to_save)
        for i in staged_to_save:
            i.save(update_fields=['transaction_model', 'updated'])

    # URL helpers
    def get_absolute_url(self) -> str:
        """Absolute URL for the receipt detail view.

        Returns
        -------
        str
            URL string for this receipt's detail page.
        """
        return reverse(
            'django_ledger:receipt-detail',
            kwargs={'entity_slug': self.entity_slug, 'receipt_pk': self.uuid},
        )

    def get_list_url(self) -> str:
        """URL for the entity's receipt list view.

        Returns
        -------
        str
            URL string for listing receipts of the same entity.
        """
        return reverse('django_ledger:receipt-list', kwargs={'entity_slug': self.entity_slug})

    def get_delete_url(self) -> str:
        """URL for the receipt delete view.

        Returns
        -------
        str
            URL string for deleting this receipt.
        """
        return reverse(
            'django_ledger:receipt-delete',
            kwargs={'entity_slug': self.entity_slug, 'receipt_pk': self.uuid},
        )

    def get_customer_list_url(self) -> Optional[str]:
        """URL for listing receipts for the same customer.

        Returns
        -------
        str or None
            URL string to the customer's receipt list, or None when the receipt
            has no customer.
        """
        if not self.customer_model_id:
            return None
        return reverse(
            'django_ledger:receipt-list-customer',
            kwargs={
                'entity_slug': self.entity_slug,
                'customer_pk': self.customer_model_id,
            },
        )

    def get_vendor_list_url(self) -> Optional[str]:
        """URL for listing receipts for the same vendor.

        Returns
        -------
        str or None
            URL string to the vendor's receipt list, or None when the receipt
            has no vendor.
        """
        if not self.vendor_model_id:
            return None
        return reverse(
            'django_ledger:receipt-list-vendor',
            kwargs={'entity_slug': self.entity_slug, 'vendor_pk': self.vendor_model_id},
        )

    def get_customer_report_url(self) -> Optional[str]:
        """URL for a report focused on this receipt's customer.

        Returns
        -------
        str or None
            URL string to the customer report, or None when the receipt has no
            customer.
        """
        if not self.customer_model_id:
            return None
        return reverse(
            'django_ledger:receipt-report-customer',
            kwargs={
                'entity_slug': self.entity_slug,
                'customer_pk': self.customer_model_id,
            },
        )

    def get_vendor_report_url(self) -> Optional[str]:
        """URL for a report focused on this receipt's vendor.

        Returns
        -------
        str or None
            URL string to the vendor report, or None when the receipt has no
            vendor.
        """
        if not self.vendor_model_id:
            return None
        return reverse(
            'django_ledger:receipt-report-vendor',
            kwargs={'entity_slug': self.entity_slug, 'vendor_pk': self.vendor_model_id},
        )

    def get_import_job_url(self) -> Optional[str]:
        """URL of the import job page associated with the staged transaction.

        Returns
        -------
        str or None
            URL string to the import job transactions page, or None when there
            is no staged transaction linked.
        """
        if not self.staged_transaction_model_id:
            return None
        job = self.staged_transaction_model.import_job
        return reverse(
            'django_ledger:import-job-detail',
            kwargs={'entity_slug': self.entity_slug, 'job_pk': job.uuid},
        )

    def get_staged_tx_url(self) -> Optional[str]:
        """Anchor URL pointing to the staged transaction within the import job page.

        Returns
        -------
        str or None
            URL string anchored to this receipt's staged transaction, or None
            when there is no staged transaction or import job URL.
        """
        if not self.staged_transaction_model_id:
            return None
        base = self.get_import_job_url()
        return f'{base}#staged-tx-{self.staged_transaction_model_id}' if base else None

    def clean(self):
        """Model validation for mutually exclusive counterparties.

        Raises
        ------
        ReceiptModelValidationError
            If a sales receipt lacks a customer or an expense receipt lacks a
            vendor.
        """
        if self.is_sales_receipt():
            if not self.customer_model_id:
                raise ReceiptModelValidationError(message=_('Sales receipt must have a customer model.'))
            self.vendor_model = None

        if self.is_expense_receipt():
            if not self.vendor_model_id:
                raise ReceiptModelValidationError(message=_('Expense receipt must have a vendor model.'))
            self.customer_model = None


class ReceiptModel(ReceiptModelAbstract):
    """Concrete receipt model ready for migration and admin usage.

    Inherits all behavior from ReceiptModelAbstract and is used as the concrete
    Django model for persistence.
    """

    class Meta:
        abstract = False


def receiptmodel_presave(instance: ReceiptModel, **kwargs):
    """Pre-save signal hook for ReceiptModel.

    Parameters
    ----------
    instance : ReceiptModel
        The model instance being saved.
    **kwargs
        Additional keyword arguments passed by Django's pre_save.

    Notes
    -----
    Currently a no-op placeholder for future pre-save logic.
    """
    pass


pre_save.connect(receiptmodel_presave, sender=ReceiptModel)



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
Miguel Sanda <msanda@arrobalytics.com>
"""

from django.core.exceptions import ValidationError


class DjangoLedgerConfigurationError(Exception):
    pass


class InvalidDateInputError(ValidationError):
    pass


class InvalidRoleError(ValidationError):
    pass


class TransactionNotInBalanceError(ValidationError):
    pass



from django.urls import path

from django_ledger import views

urlpatterns = [
    path('<slug:entity_slug>/jobs/', views.ImportJobModelListView.as_view(), name='import-job-list'),
    path(
        '<slug:entity_slug>/jobs/<uuid:job_pk>/detail/', views.ImportJobDetailView.as_view(), name='import-job-detail'
    ),
    path(
        '<slug:entity_slug>/jobs/<uuid:job_pk>/update/',
        views.ImportJobModelUpdateView.as_view(),
        name='import-job-update',
    ),
    path(
        '<slug:entity_slug>/jobs/<uuid:job_pk>/delete/',
        views.ImportJobModelDeleteView.as_view(),
        name='import-job-delete',
    ),
    path('<slug:entity_slug>/import-ofx/', views.ImportJobModelCreateView.as_view(), name='data-import-ofx'),
    path(
        '<slug:entity_slug>/jobs/<uuid:job_pk>/reset/',
        views.ImportJobModelResetView.as_view(),
        name='data-import-job-txs-undo',
    ),
    path(
        '<slug:entity_slug>/jobs/<uuid:job_pk>/txs/<uuid:staged_tx_pk>/update/',
        views.StagedTransactionUpdateView.as_view(),
        name='data-import-staged-tx-update',
    ),
    path(
        '<slug:entity_slug>/jobs/<uuid:job_pk>/txs/<uuid:staged_tx_pk>/undo/',
        views.StagedTransactionUndoView.as_view(),
        name='data-import-staged-tx-undo',
    ),
    path(
        '<slug:entity_slug>/jobs/<uuid:job_pk>/txs/<uuid:staged_tx_pk>/unmatch/',
        views.StagedTransactionUnmatchView.as_view(),
        name='data-import-staged-tx-unmatch',
    ),
]



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
    * Miguel Sanda <msanda@arrobalytics.com>
"""

from django.urls import path

from django_ledger import views

urlpatterns = [
    path(
        '<slug:entity_slug>/latest/',
        views.ReceiptModelListView.as_view(),
        name='receipt-list',
    ),
    path(
        '<slug:entity_slug>/year/<int:year>/',
        views.ReceiptModelYearListView.as_view(),
        name='receipt-list-year',
    ),
    path(
        '<slug:entity_slug>/quarter/<int:year>/q<int:quarter>/',
        views.ReceiptModelQuarterListView.as_view(),
        name='receipt-list-quarter',
    ),
    path(
        '<slug:entity_slug>/month/<int:year>/<int:month>/',
        views.ReceiptModelMonthListView.as_view(),
        name='receipt-list-month',
    ),
    path(
        '<slug:entity_slug>/detail/<uuid:receipt_pk>/',
        views.ReceiptModelDetailView.as_view(),
        name='receipt-detail',
    ),
    path(
        '<slug:entity_slug>/delete/<uuid:receipt_pk>/',
        views.ReceiptModelDeleteView.as_view(),
        name='receipt-delete',
    ),
    # Filtered lists
    path(
        '<slug:entity_slug>/type/<str:receipt_type>/',
        views.ReceiptModelListView.as_view(),
        name='receipt-list-type',
    ),
    path(
        '<slug:entity_slug>/vendor/<uuid:vendor_pk>/',
        views.ReceiptModelListView.as_view(),
        name='receipt-list-vendor',
    ),
    path(
        '<slug:entity_slug>/customer/<uuid:customer_pk>/',
        views.ReceiptModelListView.as_view(),
        name='receipt-list-customer',
    ),
    # Reports: Vendor
    path(
        '<slug:entity_slug>/report/vendor/<uuid:vendor_pk>/latest/',
        views.VendorReceiptReportListView.as_view(),
        name='receipt-report-vendor',
    ),
    path(
        '<slug:entity_slug>/report/vendor/<uuid:vendor_pk>/year/<int:year>/',
        views.VendorReceiptReportYearListView.as_view(),
        name='receipt-report-vendor-year',
    ),
    path(
        '<slug:entity_slug>/report/vendor/<uuid:vendor_pk>/quarter/<int:year>/q<int:quarter>/',
        views.VendorReceiptReportQuarterListView.as_view(),
        name='receipt-report-vendor-quarter',
    ),
    path(
        '<slug:entity_slug>/report/vendor/<uuid:vendor_pk>/month/<int:year>/<int:month>/',
        views.VendorReceiptReportMonthListView.as_view(),
        name='receipt-report-vendor-month',
    ),
    # Reports: Customer
    path(
        '<slug:entity_slug>/report/customer/<uuid:customer_pk>/latest/',
        views.CustomerReceiptReportListView.as_view(),
        name='receipt-report-customer',
    ),
    path(
        '<slug:entity_slug>/report/customer/<uuid:customer_pk>/year/<int:year>/',
        views.CustomerReceiptReportYearListView.as_view(),
        name='receipt-report-customer-year',
    ),
    path(
        '<slug:entity_slug>/report/customer/<uuid:customer_pk>/quarter/<int:year>/q<int:quarter>/',
        views.CustomerReceiptReportQuarterListView.as_view(),
        name='receipt-report-customer-quarter',
    ),
    path(
        '<slug:entity_slug>/report/customer/<uuid:customer_pk>/month/<int:year>/<int:month>/',
        views.CustomerReceiptReportMonthListView.as_view(),
        name='receipt-report-customer-month',
    ),
]



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.
"""

from django.contrib import messages
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.db.models import Q
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.translation import gettext as _
from django.views.generic import ListView, UpdateView, CreateView, DetailView
from django.views.generic import RedirectView
from django.views.generic.detail import SingleObjectMixin

from django_ledger.forms.account import AccountModelUpdateForm, AccountModelCreateForm
from django_ledger.io.io_core import get_localdate
from django_ledger.models import EntityModel, ChartOfAccountModel
from django_ledger.models.accounts import AccountModel
from django_ledger.views.mixins import (
    YearlyReportMixIn, MonthlyReportMixIn, QuarterlyReportMixIn, DjangoLedgerSecurityMixIn,
    BaseDateNavigationUrlMixIn, EntityUnitMixIn, DateReportMixIn
)


class BaseAccountModelBaseView(DjangoLedgerSecurityMixIn):
    queryset = None
    coa_model = None

    def get_authorized_entity_queryset(self):
        qs = super().get_authorized_entity_queryset()
        return qs.select_related('admin', 'default_coa', 'default_coa__entity')

    def get_coa_model(self):
        if not self.coa_model:
            entity_model: EntityModel = self.get_authorized_entity_instance()
            self.coa_model = entity_model.chartofaccountmodel_set.get(slug__exact=self.kwargs['coa_slug'])
        return self.coa_model

    def get_queryset(self):
        if self.queryset is None:
            entity_model: EntityModel = self.get_authorized_entity_instance()
            coa_slug = self.kwargs['coa_slug']

            coa_model, account_model_qs = entity_model.get_coa_accounts(
                coa_model=entity_model.default_coa if coa_slug == entity_model.default_coa_slug else coa_slug,
                return_coa_model=True,
                active=False
            )

            search = self.request.GET.get('q')
            if search:
                account_model_qs = account_model_qs.filter(
                    Q(code__icontains=search) |
                    Q(name__icontains=search)
                )

            account_model_qs = account_model_qs.select_related(
                'coa_model',
                'coa_model__entity'
            ).order_by(
                'coa_model', 'role', 'code'
            ).not_coa_root()

            self.coa_model = coa_model
            self.queryset = account_model_qs

        return super().get_queryset()

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['coa_model'] = self.get_coa_model()
        return context


# Account Views ----
class AccountModelListView(BaseAccountModelBaseView, ListView):
    template_name = 'django_ledger/account/account_list.html'
    context_object_name = 'accounts'
    PAGE_TITLE = _('Entity Accounts')
    extra_context = {
        'page_title': PAGE_TITLE,
        'header_title': PAGE_TITLE
    }
    active_only = False

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        coa_model: ChartOfAccountModel = self.get_coa_model()
        context['page_title'] = f'{coa_model.name} Accounts'
        context['header_title'] = f'{coa_model.name} Accounts'
        context['header_subtitle'] = self.get_authorized_entity_instance_name()
        context['header_subtitle_icon'] = 'ic:twotone-account-tree'
        return context

    def get_queryset(self):
        qs = super().get_queryset()
        if self.active_only:
            qs = qs.active()
        return qs

    def get(self, request, *args, **kwargs):
        response = super().get(request, *args, **kwargs)
        chart_of_accounts_model: ChartOfAccountModel = self.get_coa_model()
        if not chart_of_accounts_model.is_active():
            messages.error(request, _('WARNING: The chart of accounts list is inactive.'), extra_tags='is-danger')
        return response


class AccountModelCreateView(BaseAccountModelBaseView, CreateView):
    template_name = 'django_ledger/account/account_create.html'
    PAGE_TITLE = _('Create Account')
    extra_context = {
        'page_title': PAGE_TITLE,
        'header_title': PAGE_TITLE,
        'header_subtitle_icon': 'ic:twotone-account-tree'

    }

    def get_form(self, form_class=None):
        return AccountModelCreateForm(
            coa_model=self.get_coa_model(),
            **self.get_form_kwargs()
        )

    def get_initial(self):
        return {
            'coa_model': self.get_coa_model(),
        }

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        coa_model = self.get_coa_model()
        context['coa_model'] = coa_model
        context['header_subtitle'] = f'CoA: {coa_model.name}'
        return context

    def form_valid(self, form: AccountModelCreateForm):
        account_model: AccountModel = form.save(commit=False)
        coa_model = account_model.coa_model
        coa_model.insert_account(account_model=account_model)
        return HttpResponseRedirect(self.get_success_url())

    def get_success_url(self):
        coa_model: ChartOfAccountModel = self.get_coa_model()
        return coa_model.get_account_list_url()


class AccountModelUpdateView(BaseAccountModelBaseView, UpdateView):
    context_object_name = 'account'
    template_name = 'django_ledger/account/account_update.html'
    slug_url_kwarg = 'account_pk'
    slug_field = 'uuid'
    form_class = AccountModelUpdateForm

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['page_title'] = _('Update Account')
        context['header_title'] = _(f'Update Account: {self.object.code} - {self.object.name}')
        context['header_subtitle_icon'] = 'ic:twotone-account-tree'
        return context

    def get_success_url(self):
        coa_model: ChartOfAccountModel = self.get_coa_model()
        return coa_model.get_account_list_url()


class AccountModelDetailView(BaseAccountModelBaseView, RedirectView):

    def get_redirect_url(self, *args, **kwargs):
        loc_date = get_localdate()
        entity_model: EntityModel = self.get_authorized_entity_instance()
        return reverse('django_ledger:account-detail-month',
                       kwargs={
                           'entity_slug': entity_model.slug,
                           'account_pk': self.kwargs['account_pk'],
                           'coa_slug': self.kwargs['coa_slug'],
                           'year': loc_date.year,
                           'month': loc_date.month,
                       })


class AccountModelYearDetailView(BaseAccountModelBaseView,
                                 BaseDateNavigationUrlMixIn,
                                 EntityUnitMixIn,
                                 YearlyReportMixIn,
                                 DetailView):
    context_object_name = 'account'
    template_name = 'django_ledger/account/account_detail.html'
    slug_url_kwarg = 'account_pk'
    slug_field = 'uuid'
    DEFAULT_TXS_DAYS = 30
    extra_context = {
        'DEFAULT_TXS_DAYS': DEFAULT_TXS_DAYS,
        'header_subtitle_icon': 'ic:round-account-tree'
    }

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        account_model: AccountModel = context['object']
        context['header_title'] = f'Account {account_model.code} - {account_model.name}'
        context['page_title'] = f'Account {account_model.code} - {account_model.name}'
        txs_qs = account_model.transactionmodel_set.all().not_closing_entry().posted().order_by(
            'journal_entry__timestamp'
        ).select_related(
            'journal_entry',
            'journal_entry__entity_unit',
            'journal_entry__ledger__billmodel',
            'journal_entry__ledger__invoicemodel',
        )
        txs_qs = txs_qs.from_date(self.get_from_date())
        txs_qs = txs_qs.to_date(self.get_to_date())
        context['transactions'] = txs_qs
        return context


class AccountModelQuarterDetailView(AccountModelYearDetailView, QuarterlyReportMixIn):
    """
    Account Model Quarter Detail View
    """


class AccountModelMonthDetailView(AccountModelYearDetailView, MonthlyReportMixIn):
    """
    Account Model Month Detail View
    """


class AccountModelDateDetailView(AccountModelYearDetailView, DateReportMixIn):
    """
    Account Model Date Detail View
    """


# ACTIONS...
class BaseAccountModelActionView(BaseAccountModelBaseView,
                                 RedirectView,
                                 SingleObjectMixin):
    http_method_names = ['get']
    pk_url_kwarg = 'account_pk'
    action_name = None
    commit = True

    def get_redirect_url(self, *args, **kwargs):
        account_model: AccountModel = self.get_object()
        return account_model.get_coa_account_list_url()

    def get(self, request, *args, **kwargs):
        kwargs['user_model'] = self.request.user
        if not self.action_name:
            raise ImproperlyConfigured('View attribute action_name is required.')
        response = super(BaseAccountModelActionView, self).get(request, *args, **kwargs)
        account_model: AccountModel = self.get_object()

        try:
            getattr(account_model, self.action_name)(commit=self.commit, **kwargs)
        except ValidationError as e:
            messages.add_message(
                request,
                message=e.message,
                level=messages.ERROR,
                extra_tags='is-danger'
            )
        return response



# Generated by Django 5.2.7 on 2025-11-01 20:38

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_ledger', '0027_alter_accountmodel_role_alter_receiptmodel_amount_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='stagedtransactionmodel',
            name='matched_transaction_model',
            field=models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='matchedtransaction_model', to='django_ledger.transactionmodel', verbose_name='Matched Transaction Model'),
        ),
        migrations.AlterField(
            model_name='stagedtransactionmodel',
            name='transaction_model',
            field=models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='django_ledger.transactionmodel', verbose_name='Transaction Model'),
        ),
        migrations.AddIndex(
            model_name='stagedtransactionmodel',
            index=models.Index(fields=['matched_transaction_model'], name='django_ledg_matched_dde2a7_idx'),
        ),
    ]



from itertools import groupby
from typing import Optional
from uuid import UUID

from django import forms
from django.db.models import Q
from django.forms import (
    BaseModelFormSet,
    CheckboxInput,
    HiddenInput,
    ModelForm,
    NumberInput,
    Select,
    Textarea,
    TextInput,
    ValidationError,
    modelformset_factory,
)
from django.utils.translation import gettext_lazy as _

from django_ledger.io import GROUP_DEBT_PAYMENT, GROUP_EXPENSES, GROUP_INCOME, GROUP_TRANSFERS
from django_ledger.models import (
    EntityModel,
    ImportJobModel,
    StagedTransactionModel,
)
from django_ledger.settings import DJANGO_LEDGER_FORM_INPUT_CLASSES


class ImportJobModelCreateForm(ModelForm):
    def __init__(self, entity_model: EntityModel, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ENTITY_MODEL: EntityModel = entity_model
        self.fields['bank_account_model'].queryset = self.ENTITY_MODEL.bankaccountmodel_set.all().active()

    ofx_file = forms.FileField(
        label='Select File...',
        required=True,
        widget=forms.FileInput(attrs={'class': 'file-input', 'accept': '.ofx,.qfx'}),
    )

    class Meta:
        model = ImportJobModel
        fields = [
            'bank_account_model',
            'description',
        ]
        widgets = {
            'description': TextInput(
                attrs={
                    'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-large',
                    'placeholder': _('Name this import...'),
                }
            ),
            'bank_account_model': Select(
                attrs={
                    'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                }
            ),
        }
        help_texts = {
            'bank_account_model': _('Select the bank account to import transactions from.'),
        }


class ImportJobModelUpdateForm(ModelForm):
    class Meta:
        model = ImportJobModel
        fields = ['description']
        widgets = {'description': TextInput(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES})}


class BaseStagedTransactionModelFormSet(BaseModelFormSet):
    def __init__(
            self,
            *args,
            entity_model: EntityModel,
            import_job_model: ImportJobModel,
            staged_tx_pk: Optional[UUID | StagedTransactionModel] = None,
            **kwargs,
    ):
        super().__init__(*args, **kwargs)

        # validates that the job import model belongs to the entity model...
        if import_job_model.entity_uuid != entity_model.uuid:
            raise ValidationError(message=_('Import job does not belong to this entity'))

        self.ENTITY_MODEL = entity_model
        self.IMPORT_JOB_MODEL: ImportJobModel = import_job_model
        self.STAGED_TX_MODEL: Optional[UUID | StagedTransactionModel] = staged_tx_pk

        staged_txs_qs = self.IMPORT_JOB_MODEL.stagedtransactionmodel_set.select_related(
            'import_job',
            'import_job__ledger_model',
            'import_job__ledger_model__entity',
            'import_job__bank_account_model__account_model',
            'vendor_model',
            'customer_model',
            'unit_model',
        ).is_pending()

        if self.STAGED_TX_MODEL:
            if isinstance(self.STAGED_TX_MODEL, StagedTransactionModel):
                staged_txs_qs = staged_txs_qs.filter(
                    Q(uuid__exact=self.STAGED_TX_MODEL.uuid) | Q(parent__uuid__exact=self.STAGED_TX_MODEL.uuid)
                )
            elif isinstance(self.STAGED_TX_MODEL, UUID):
                staged_txs_qs = staged_txs_qs.filter(
                    Q(uuid__exact=self.STAGED_TX_MODEL) | Q(parent__uuid__exact=self.STAGED_TX_MODEL)
                )
            else:
                staged_txs_qs = staged_txs_qs.none()

        self.queryset = staged_txs_qs

        self.MAPPED_ACCOUNT_MODEL = self.IMPORT_JOB_MODEL.bank_account_model.account_model

        self.account_model_qs = (
            self.ENTITY_MODEL.get_coa_accounts().available().exclude(uuid__exact=self.MAPPED_ACCOUNT_MODEL.uuid)
        )
        self.ACCOUNT_MODEL_CHOICES = [(None, '----')] + [(a.uuid, a) for a in self.account_model_qs]
        self.ACCOUNT_MODEL_EXPENSES_CHOICES = [(None, '----')] + [
            (a.uuid, a) for a in self.account_model_qs if a.role in GROUP_EXPENSES
        ]
        self.ACCOUNT_MODEL_SALES_CHOICES = [(None, '----')] + [
            (a.uuid, a) for a in self.account_model_qs if a.role in GROUP_INCOME
        ]

        self.ACCOUNT_MODEL_TRANSFERS_CHOICES = [(None, '----')] + [
            (a.uuid, a) for a in self.account_model_qs if a.role in GROUP_TRANSFERS
        ]

        self.ACCOUNT_MODEL_DEBT_PAYMENT_CHOICES = [(None, '----')] + [
            (a.uuid, a) for a in self.account_model_qs if a.role in GROUP_DEBT_PAYMENT
        ]

        self.unit_model_qs = entity_model.entityunitmodel_set.all()
        self.UNIT_MODEL_CHOICES = [(None, '----')] + [(u.uuid, u) for i, u in enumerate(self.unit_model_qs)]

        self.VENDOR_MODEL_QS = entity_model.vendormodel_set.visible().order_by('vendor_name')
        self.CUSTOMER_MODEL_QS = entity_model.customermodel_set.visible().order_by('customer_name')

        self.VENDOR_CHOICES = [(None, '-----')] + [(str(v.uuid), v) for v in self.VENDOR_MODEL_QS]
        self.CUSTOMER_CHOICES = [(None, '-----')] + [(str(c.uuid), c) for c in self.CUSTOMER_MODEL_QS]

        self.VENDOR_MAP = dict(self.VENDOR_CHOICES)
        self.CUSTOMER_MAP = dict(self.CUSTOMER_CHOICES)

        self.FORMS_BY_ID = {
            f.instance.uuid: f for f in self.forms if getattr(f, 'instance', None) and getattr(f.instance, 'uuid', None)
        }

        form_children = [(f.instance.parent_id, f.instance.uuid) for f in self.forms if f.instance.parent_id]
        form_children.sort(key=lambda f: f[0])

        self.FORM_CHILDREN = {g: list(j[1] for j in p) for g, p in groupby(form_children, key=lambda i: i[0])}

    def get_form_kwargs(self, index):
        return {
            'base_formset_instance': self,
            'entity_model': self.ENTITY_MODEL,
            'import_job_model': self.IMPORT_JOB_MODEL,
        }


class StagedTransactionModelForm(ModelForm):
    tx_import = forms.BooleanField(
        initial=False,
        required=False,
        widget=forms.CheckboxInput(
            attrs={
                'class': 'is-checkradio',
            }
        ),
        label=_('Import Transaction'),
        help_text=_('Will import this transaction when saved...'),
    )

    tx_split = forms.BooleanField(
        initial=False,
        required=False,
        widget=forms.CheckboxInput(
            attrs={
                'class': 'is-checkradio',
            }
        ),
        label=_('Add a Transaction Split?'),
        help_text=_('Will split this transaction when saved...'),
    )

    account_search = forms.CharField(
        required=False,
        widget=forms.TextInput(
            attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-small',
                'data-account-search': None,
                'placeholder': _('Search Account...'),
            }
        ),
        label=_('Search Account...'),
        help_text=_('Filters the account list...'),
    )

    def __init__(
            self,
            *args,
            entity_model: EntityModel,
            import_job_model: ImportJobModel,
            base_formset_instance: BaseStagedTransactionModelFormSet,
            **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.ENTITY_MODEL: EntityModel = entity_model
        self.IMPORT_JOB_MODEL: ImportJobModel = import_job_model
        self.BASE_FORMSET: BaseStagedTransactionModelFormSet = base_formset_instance

        self.ACCOUNT_MODEL_CHOICES = self.BASE_FORMSET.ACCOUNT_MODEL_CHOICES
        self.EXPENSE_ACCOUNT_CHOICES = self.BASE_FORMSET.ACCOUNT_MODEL_EXPENSES_CHOICES
        self.SALES_ACCOUNT_CHOICES = self.BASE_FORMSET.ACCOUNT_MODEL_SALES_CHOICES
        self.TRANSFER_ACCOUNT_CHOICES = self.BASE_FORMSET.ACCOUNT_MODEL_TRANSFERS_CHOICES
        self.DEBT_PAYDOWN_ACCOUNT_CHOICES = self.BASE_FORMSET.ACCOUNT_MODEL_DEBT_PAYMENT_CHOICES
        self.UNIT_MODEL_CHOICES = self.BASE_FORMSET.UNIT_MODEL_CHOICES
        self.VENDOR_CHOICES = self.BASE_FORMSET.VENDOR_CHOICES
        self.CUSTOMER_CHOICES = self.BASE_FORMSET.CUSTOMER_CHOICES

        self.VENDOR_MAP = dict(self.VENDOR_CHOICES)
        self.CUSTOMER_MAP = dict(self.CUSTOMER_CHOICES)

        self.fields['vendor_model'].choices = self.VENDOR_CHOICES
        self.fields['customer_model'].choices = self.CUSTOMER_CHOICES
        self.fields['account_model'].choices = self.ACCOUNT_MODEL_CHOICES
        self.fields['unit_model'].choices = self.UNIT_MODEL_CHOICES

        self.fields['activity'].disabled = True

        staged_tx_model: StagedTransactionModel = getattr(self, 'instance', None)

        if staged_tx_model:
            if not staged_tx_model.can_have_activity():
                self.fields['activity'].widget = HiddenInput()

            if staged_tx_model.is_sales():
                self.fields['account_model'].choices = self.SALES_ACCOUNT_CHOICES
            if staged_tx_model.is_expense():
                self.fields['account_model'].choices = self.EXPENSE_ACCOUNT_CHOICES
            if staged_tx_model.is_transfer():
                self.fields['account_model'].choices = self.TRANSFER_ACCOUNT_CHOICES
            if staged_tx_model.is_debt_payment():
                self.fields['account_model'].choices = self.DEBT_PAYDOWN_ACCOUNT_CHOICES

            if not staged_tx_model.can_match():
                self.fields['matched_transaction'].widget = HiddenInput()
                self.fields['matched_transaction'].disabled = True

            candidates_qs = staged_tx_model.get_match_candidates_qs()

            if any([staged_tx_model.is_transfer(), staged_tx_model.is_debt_payment()]) and candidates_qs.exists():
                self.fields['matched_transaction_model'].queryset = candidates_qs
                if candidates_qs.count() == 1:
                    self.fields['matched_transaction_model'].initial = candidates_qs.first()
            else:
                self.fields['matched_transaction_model'].widget = HiddenInput()
                self.fields['matched_transaction_model'].disabled = True

            if staged_tx_model.has_match():
                self.fields['account_model'].disabled = True
                self.fields['account_search'].disabled = True

                self.fields['tx_split'].widget = HiddenInput()
                self.fields['tx_split'].disabled = True

                self.fields['vendor_model'].disabled = False
                self.fields['vendor_model'].widget = HiddenInput()
                self.fields['customer_model'].disabled = False
                self.fields['customer_model'].widget = HiddenInput()
            else:
                self.fields['matched_transaction'].disabled = False
                self.fields['matched_transaction'].widget = HiddenInput()

            if not staged_tx_model.has_children():
                self.fields['bundle_split'].widget = HiddenInput()
                self.fields['bundle_split'].disabled = True

            if not staged_tx_model.can_have_amount_split():
                self.fields['amount_split'].widget = HiddenInput()
                self.fields['amount_split'].disabled = True

            if not staged_tx_model.can_have_bundle_split():
                self.fields['bundle_split'].widget = HiddenInput()
                self.fields['bundle_split'].disabled = True

            if not staged_tx_model.can_have_receipt():
                self.fields['receipt_type'].widget = HiddenInput()
                self.fields['receipt_type'].disabled = True

            if not staged_tx_model.can_have_account():
                self.fields['account_model'].widget = HiddenInput()
                self.fields['account_model'].disabled = True

                self.fields['account_search'].widget = HiddenInput()
                self.fields['account_search'].disabled = True

            if not staged_tx_model.can_have_unit():
                self.fields['unit_model'].widget = HiddenInput()
                self.fields['unit_model'].disabled = True

            if not staged_tx_model.can_have_vendor():
                self.fields['vendor_model'].widget = HiddenInput()
                self.fields['vendor_model'].disabled = True

            if not staged_tx_model.can_have_customer():
                self.fields['customer_model'].widget = HiddenInput()
                self.fields['customer_model'].disabled = True

            if not staged_tx_model.can_import():
                self.fields['tx_import'].widget = HiddenInput()
                self.fields['tx_import'].disabled = True

            if not staged_tx_model.can_split():
                self.fields['tx_split'].widget = HiddenInput()
                self.fields['tx_split'].disabled = True

            if staged_tx_model.is_children():
                if 'tx_import' in self.fields:
                    self.fields['tx_import'].widget = HiddenInput()
                    self.fields['tx_import'].disabled = True
                if 'tx_split' in self.fields:
                    self.fields['tx_split'].widget = HiddenInput()
                    self.fields['tx_split'].disabled = True

    def clean_account_model(self):
        staged_txs_model: StagedTransactionModel = self.instance
        if staged_txs_model.has_children():
            return None
        return self.cleaned_data['account_model']

    def clean_unit_model(self):
        staged_txs_model: StagedTransactionModel = self.instance
        if not staged_txs_model.can_have_unit():
            if staged_txs_model.is_children():
                return staged_txs_model.parent.unit_model
            return None
        return self.cleaned_data['unit_model']

    def clean_bundle_split(self):
        staged_txs_model: StagedTransactionModel = self.instance
        if staged_txs_model.is_single():
            return True
        if staged_txs_model.is_children():
            return staged_txs_model.parent.bundle_split
        return self.cleaned_data['bundle_split']

    def clean_activity(self):
        return self.instance.activity

    def clean(self):
        # Prevent conflicting actions
        if self.cleaned_data.get('tx_import') and self.cleaned_data.get('tx_split'):
            raise ValidationError(message=_('Cannot import and split at the same time'))

        staged_txs_model: StagedTransactionModel = self.instance

        # Matching validation for transfers & debt payments
        if all(
                [
                    self.cleaned_data.get('matched_transaction') is True,
                    any([staged_txs_model.is_transfer(), staged_txs_model.is_debt_payment()]),
                ]
        ):
            candidates_qs = staged_txs_model.get_match_candidates_qs()
            selected = self.cleaned_data.get('matched_transaction_model')
            if candidates_qs.count() > 1 and selected is None:
                raise ValidationError(message=_('Multiple matches found. Please select a transaction to match.'))
            if candidates_qs.count() == 1 and selected is None:
                self.cleaned_data['matched_transaction_model'] = candidates_qs.first()

    class Meta:
        model = StagedTransactionModel
        fields = [
            'activity',
            'matched_transaction_model',
            'account_search',
            'account_model',
            'amount_split',
            'unit_model',
            'receipt_type',
            'vendor_model',
            'customer_model',
            'notes',
            'bundle_split',
            'matched_transaction',
            'tx_import',
        ]
        labels = {
            'bundle_split': _('Grouped Transactions'),
            'unit_model': _('Assigned Unit or Class'),
            'amount_split': _('Split Amount'),
            'account_model': _('Map to Account'),
            'matched_transaction': _('Match Transaction'),
            'vendor_model': _('Vendor'),
            'customer_model': _('Customer'),
        }
        help_texts = {
            'bundle_split': _('If checked, maps all splits to the same unit & receipt.'),
            'amount_split': _('All splits amount must total parent transaction amount.'),
            'account_model': _('The CoA account where the transaction will be imported.'),
            'matched_transaction': _('The transaction will be matched against the selected transaction.'),
        }
        widgets = {
            'account_model': Select(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES}),
            'unit_model': Select(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES}),
            'amount_split': NumberInput(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES}),
            'vendor_model': Select(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES}),
            'customer_model': Select(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES}),
            'receipt_type': Select(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES}),
            'matched_transaction_model': Select(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES}),
            'notes': Textarea(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES}),
            'activity': Select(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES}),
            'bundle_split': CheckboxInput(attrs={'class': 'is-checkradio'}),
        }


StagedTransactionModelFormSet = modelformset_factory(
    model=StagedTransactionModel,
    form=StagedTransactionModelForm,
    formset=BaseStagedTransactionModelFormSet,
    can_delete=True,
    can_order=False,
    extra=0,
)



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
    * Miguel Sanda <msanda@arrobalytics.com>
"""

from typing import Optional
from calendar import month_name

from django.http import HttpResponseForbidden
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext_lazy as _
from django.views.generic import (
    ArchiveIndexView,
    DeleteView,
    DetailView,
    MonthArchiveView,
    YearArchiveView,
)

from django_ledger.models import CustomerModel, EntityModel, LedgerModel, VendorModel
from django_ledger.models.receipt import ReceiptModel, ReceiptModelQuerySet
from django_ledger.models.transactions import TransactionModel
from django_ledger.views.mixins import (
    DjangoLedgerSecurityMixIn,
    QuarterlyReportMixIn,
)


class BaseReceiptModelViewMixIn(DjangoLedgerSecurityMixIn):
    queryset: Optional[ReceiptModelQuerySet] = None

    def get_queryset(self):
        if self.queryset is None:
            entity_model: EntityModel = self.AUTHORIZED_ENTITY_MODEL
            qs = entity_model.get_receipts()
            qs = qs.select_related(
                'ledger_model', 'customer_model', 'vendor_model'
            ).order_by('-receipt_date', '-created')

            receipt_type = self.kwargs.get('receipt_type')
            if receipt_type:
                qs = qs.filter(receipt_type__exact=receipt_type)
                if receipt_type in [
                    ReceiptModel.SALES_RECEIPT,
                    ReceiptModel.SALES_REFUND,
                ]:
                    qs = qs.filter(
                        customer_model__isnull=False, vendor_model__isnull=True
                    )

                elif receipt_type in [
                    ReceiptModel.EXPENSE_RECEIPT,
                    ReceiptModel.EXPENSE_REFUND,
                ]:
                    qs = qs.filter(
                        vendor_model__isnull=False, customer_model__isnull=True
                    )

            vendor_pk = self.kwargs.get('vendor_pk')
            if vendor_pk:
                qs = qs.for_vendor(vendor_model=vendor_pk)

            customer_pk = self.kwargs.get('customer_pk')
            if customer_pk:
                qs = qs.for_customer(customer_model=customer_pk)

            self.queryset = qs

        return self.queryset


class ReceiptModelListView(BaseReceiptModelViewMixIn, ArchiveIndexView):
    template_name = 'django_ledger/receipt/receipt_list.html'
    context_object_name = 'receipt_list'
    PAGE_TITLE = _('Receipts List')
    date_field = 'receipt_date'
    paginate_by = 20
    paginate_orphans = 2
    allow_empty = True
    extra_context = {
        'title': PAGE_TITLE,
        'page_title': PAGE_TITLE,
        'header_title': PAGE_TITLE,
        'header_subtitle_icon': 'mdi:receipt',
    }

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        subtitle = None

        receipt_type = self.kwargs.get('receipt_type')

        if receipt_type:
            context['title'] = ReceiptModel.RECEIPT_TYPES_MAP[receipt_type]

        vendor_pk = self.kwargs.get('vendor_pk')
        if vendor_pk:
            vendor = VendorModel.objects.for_entity(
                entity_model=self.AUTHORIZED_ENTITY_MODEL
            ).get(uuid__exact=vendor_pk)
            subtitle = vendor.vendor_name
        customer_pk = self.kwargs.get('customer_pk')

        if customer_pk:
            customer = CustomerModel.objects.for_entity(
                entity_model=self.AUTHORIZED_ENTITY_MODEL
            ).get(uuid__exact=customer_pk)
            subtitle = customer.customer_name

        if subtitle:
            context['header_subtitle'] = subtitle
        return context


class ReceiptModelYearListView(ReceiptModelListView, YearArchiveView):
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['year'] = self.get_year()
        context['page_title'] = _(f'Receipts List {self.get_year()}')
        context['header_title'] = _(f'Receipts List {self.get_year()}')
        context['header_subtitle'] = self.AUTHORIZED_ENTITY_MODEL.name
        return context


class ReceiptModelQuarterListView(ReceiptModelYearListView, QuarterlyReportMixIn):
    def get_queryset(self):
        qs = super().get_queryset()
        return qs.for_dates(from_date=self.get_from_date(), to_date=self.get_to_date())

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['page_title'] = _(f'Receipts List Q{self.get_quarter()}')
        context['header_title'] = _(f'Receipts List Q{self.get_quarter()}')
        context['header_subtitle'] = self.AUTHORIZED_ENTITY_MODEL.name
        return context


class ReceiptModelMonthListView(ReceiptModelYearListView, MonthArchiveView):
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        year = self.get_year()
        month_num = int(self.get_month())
        month_label = month_name[month_num]
        context['page_title'] = _(f'Receipts List {month_label} {year}')
        context['header_title'] = _(f'Receipts List {month_label}, {year}')
        context['header_subtitle'] = self.AUTHORIZED_ENTITY_MODEL.name
        return context


class ReceiptModelDetailView(BaseReceiptModelViewMixIn, DetailView):
    template_name = 'django_ledger/receipt/receipt_detail.html'
    context_object_name = 'receipt'
    slug_field = 'uuid'
    slug_url_kwarg = 'receipt_pk'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        receipt_model: ReceiptModel = self.object
        ledger_model: LedgerModel = receipt_model.ledger_model
        title = _(f'Receipt {receipt_model.receipt_number}')
        context['page_title'] = title
        context['header_title'] = title
        context['header_subtitle'] = receipt_model.receipt_date
        context['header_subtitle_icon'] = 'mdi:receipt'

        tx_list = (
            TransactionModel.objects.for_entity(
                entity_model=self.AUTHORIZED_ENTITY_MODEL
            )
            .for_ledger(ledger_model=ledger_model)
            .posted()
            .not_closing_entry()
            .select_related(
                'account',
                'journal_entry',
                'journal_entry__entity_unit',
            )
            .order_by('journal_entry__timestamp', 'account__code')
        )

        context['tx_list'] = tx_list
        context['staged_tx'] = receipt_model.staged_transaction_model
        if receipt_model.staged_transaction_model_id:
            context['import_job'] = receipt_model.staged_transaction_model.import_job
        return context


# VENDOR VIEWS......
class VendorReceiptReportListView(ReceiptModelListView):
    template_name = 'django_ledger/receipt/vendor_receipt_report.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        vendor_pk = self.kwargs['vendor_pk']
        vendor_model_qs = VendorModel.objects.for_entity(
            entity_model=self.AUTHORIZED_ENTITY_MODEL
        )
        vendor_model: VendorModel = get_object_or_404(
            vendor_model_qs, uuid__exact=vendor_pk
        )
        context['vendor_model'] = vendor_model
        context['page_title'] = _(f'Vendor Receipts {vendor_model.vendor_name}')
        context['header_title'] = _('Vendor Receipts')
        context['header_subtitle'] = vendor_model.vendor_name
        context['header_subtitle_icon'] = 'mdi:receipt'
        return context


class VendorReceiptReportYearListView(ReceiptModelYearListView):
    template_name = 'django_ledger/receipt/vendor_receipt_report.html'


class VendorReceiptReportQuarterListView(ReceiptModelQuarterListView):
    template_name = 'django_ledger/receipt/vendor_receipt_report.html'


class VendorReceiptReportMonthListView(ReceiptModelMonthListView):
    template_name = 'django_ledger/receipt/vendor_receipt_report.html'


# CUSTOMERS VIEWS......
class CustomerReceiptReportListView(ReceiptModelListView):
    template_name = 'django_ledger/receipt/customer_receipt_report.html'
    allow_empty = True

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        customer_model_qs = CustomerModel.objects.for_entity(
            entity_model=self.AUTHORIZED_ENTITY_MODEL
        )
        customer_pk = self.kwargs['customer_pk']
        customer_model = get_object_or_404(
            customer_model_qs,
            uuid__exact=customer_pk,
        )
        context['vendor_model'] = customer_model
        context['page_title'] = _(f'Customer Receipts {customer_model.name}')
        context['header_title'] = _('Customer Receipts')
        context['header_subtitle'] = customer_model.vendor_name
        context['header_subtitle_icon'] = 'mdi:receipt'
        return context


class CustomerReceiptReportYearListView(CustomerReceiptReportListView):
    template_name = 'django_ledger/receipt/customer_receipt_report.html'
    make_object_list = True


class CustomerReceiptReportQuarterListView(ReceiptModelQuarterListView):
    template_name = 'django_ledger/receipt/customer_receipt_report.html'
    make_object_list = True


class CustomerReceiptReportMonthListView(ReceiptModelMonthListView):
    template_name = 'django_ledger/receipt/customer_receipt_report.html'
    month_format = '%m'


class ReceiptModelDeleteView(BaseReceiptModelViewMixIn, DeleteView):
    template_name = 'django_ledger/receipt/receipt_delete.html'
    context_object_name = 'receipt'
    slug_field = 'uuid'
    slug_url_kwarg = 'receipt_pk'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        receipt: ReceiptModel = self.object
        title = _(f'Delete Receipt {receipt.receipt_number}')
        context['page_title'] = title
        context['header_title'] = title
        context['header_subtitle_icon'] = 'mdi:receipt'
        return context

    def can_delete(self, receipt_model: ReceiptModel) -> bool:
        entity_model: EntityModel = self.AUTHORIZED_ENTITY_MODEL
        ce_date = entity_model.get_closing_entry_for_date(
            io_date=receipt_model.receipt_date, inclusive=True
        )
        return ce_date is None

    def delete(self, request, *args, **kwargs):
        receipt_model: ReceiptModel = self.object
        if not receipt_model.can_delete():
            return HttpResponseForbidden(
                'Receipt cannot be deleted because it falls within a closed period.'
            )
        return super().delete(request, *args, **kwargs)

    def get_success_url(self):
        receipt_model: ReceiptModel = self.object
        return receipt_model.get_list_url()



# Generated by Django 5.2.6 on 2025-09-27 22:04

import django.core.validators
import django.db.models.deletion
import django_ledger.io.io_core
import uuid
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_ledger', '0025_alter_billmodel_cash_account_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='stagedtransactionmodel',
            name='customer_model',
            field=models.ForeignKey(blank=True, help_text='The Customer associated with the transaction.', null=True, on_delete=django.db.models.deletion.RESTRICT, to='django_ledger.customermodel', verbose_name='Associated Customer Model'),
        ),
        migrations.AddField(
            model_name='stagedtransactionmodel',
            name='receipt_type',
            field=models.CharField(blank=True, choices=[('sales', 'Sales Receipt'), ('customer_refund', 'Sales Refund'), ('expense', 'Expense Receipt'), ('expense_refund', 'Expense Refund'), ('transfer', 'Transfer Receipt')], help_text='The receipt type of the transaction.', max_length=20, null=True, verbose_name='Receipt Type'),
        ),
        migrations.AddField(
            model_name='stagedtransactionmodel',
            name='vendor_model',
            field=models.ForeignKey(blank=True, help_text='The Vendor associated with the transaction.', null=True, on_delete=django.db.models.deletion.RESTRICT, to='django_ledger.vendormodel', verbose_name='Associated Vendor Model'),
        ),
        migrations.CreateModel(
            name='ReceiptModel',
            fields=[
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, null=True)),
                ('markdown_notes', models.TextField(blank=True, null=True, verbose_name='Markdown Notes')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('receipt_number', models.CharField(max_length=255, verbose_name='Receipt Number')),
                ('receipt_date', models.DateField(verbose_name='Receipt Date')),
                ('receipt_type', models.CharField(choices=[('sales', 'Sales Receipt'), ('customer_refund', 'Sales Refund'), ('expense', 'Expense Receipt'), ('expense_refund', 'Expense Refund'), ('transfer', 'Transfer Receipt')], verbose_name='Receipt Type')),
                ('amount', models.DecimalField(decimal_places=2, help_text='Amount of the receipt.', max_digits=20, validators=[django.core.validators.MinValueValidator(0)], verbose_name='Receipt Amount')),
                ('charge_account', models.ForeignKey(help_text='The financial account (cash or credit) where this transaction was made.', on_delete=django.db.models.deletion.PROTECT, related_name='charge_receiptmodel_set', to='django_ledger.accountmodel', verbose_name='Charge Account')),
                ('customer_model', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='django_ledger.customermodel', verbose_name='Customer Model')),
                ('ledger_model', models.ForeignKey(editable=False, on_delete=django.db.models.deletion.PROTECT, to='django_ledger.ledgermodel', verbose_name='Ledger Model')),
                ('receipt_account', models.ForeignKey(help_text='The income or expense account where this transaction will be reflected', on_delete=django.db.models.deletion.PROTECT, to='django_ledger.accountmodel', verbose_name='PnL Account')),
                ('staged_transaction_model', models.OneToOneField(blank=True, help_text='The staged transaction associated with the receipt from bank feeds.', null=True, on_delete=django.db.models.deletion.RESTRICT, to='django_ledger.stagedtransactionmodel', verbose_name='Staged Transaction Model')),
                ('unit_model', models.ForeignKey(blank=True, help_text='Helps segregate receipts and transactions into different classes or departments.', null=True, on_delete=django.db.models.deletion.PROTECT, to='django_ledger.entityunitmodel', verbose_name='Unit Model')),
                ('vendor_model', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='django_ledger.vendormodel', verbose_name='Vendor Model')),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model, django_ledger.io.io_core.IOMixIn),
        ),
    ]



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
    * Miguel Sanda <msanda@arrobalytics.com>
"""

from datetime import datetime, time
from typing import Optional

from django.contrib import messages
from django.db import transaction
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse
from django.utils.timezone import make_aware
from django.utils.translation import gettext_lazy as _
from django.views import View
from django.views.generic import DeleteView, DetailView, FormView, ListView, UpdateView
from django_ledger.forms.data_import import (
    ImportJobModelCreateForm,
    ImportJobModelUpdateForm,
    StagedTransactionModelFormSet,
)
from django_ledger.io.ofx import OFXFileManager
from django_ledger.models import (
    StagedTransactionModelValidationError,
)
from django_ledger.models.data_import import ImportJobModel, StagedTransactionModel
from django_ledger.views.mixins import DjangoLedgerSecurityMixIn


# Import Job Views....
class ImportJobModelViewBaseView(DjangoLedgerSecurityMixIn):
    queryset = None

    def get_queryset(self):
        if self.queryset is None:
            self.queryset = (
                ImportJobModel.objects.for_entity(
                    entity_model=self.AUTHORIZED_ENTITY_MODEL,
                )
                .order_by('-created')
                .select_related(
                    'bank_account_model',
                    'bank_account_model__entity_model',
                    'bank_account_model__account_model',
                    'bank_account_model__account_model__coa_model',
                )
            )
        return self.queryset


class ImportJobModelCreateView(ImportJobModelViewBaseView, FormView):
    template_name = 'django_ledger/data_import/import_job_create.html'
    PAGE_TITLE = _('Create Import Job')
    extra_context = {'page_title': PAGE_TITLE, 'header_title': PAGE_TITLE}
    form_class = ImportJobModelCreateForm

    def get_form(self, form_class=None, **kwargs):
        return self.form_class(entity_model=self.get_authorized_entity_instance(), **self.get_form_kwargs())

    def get_success_url(self):
        return reverse(
            'django_ledger:import-job-list',
            kwargs={'entity_slug': self.AUTHORIZED_ENTITY_MODEL.slug},
        )

    def form_valid(self, form):
        ofx_manager = OFXFileManager(ofx_file_or_path=form.files['ofx_file'])
        import_job: ImportJobModel = form.save(commit=False)
        import_job.configure(commit=False)
        import_job.save()

        txs_to_stage = ofx_manager.get_account_txs()
        staged_txs_model_list = [
            StagedTransactionModel(
                date_posted=make_aware(value=datetime.combine(date=tx.dtposted.date(), time=time.min)),
                fit_id=tx.fitid,
                amount=tx.trnamt,
                import_job=import_job,
                name=tx.name,
                memo=tx.memo,
            )
            for tx in txs_to_stage
        ]
        for tx in staged_txs_model_list:
            tx.clean()

        StagedTransactionModel.objects.bulk_create(staged_txs_model_list)
        return super().form_valid(form=form)


class ImportJobModelListView(ImportJobModelViewBaseView, ListView):
    template_name = 'django_ledger/data_import/import_job_list.html'
    PAGE_TITLE = _('Data Import Jobs')
    extra_context = {'page_title': PAGE_TITLE, 'header_title': PAGE_TITLE}
    context_object_name = 'import_job_list'


class ImportJobDetailView(ImportJobModelViewBaseView, DetailView):
    template_name = 'django_ledger/data_import/import_job_detail.html'
    PAGE_TITLE = _('Import Job Detail')
    context_object_name = 'import_job_model'
    pk_url_kwarg = 'job_pk'
    http_method_names = ['get']

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        import_job_model: ImportJobModel = self.object

        # Base queryset with all staged transactions for this job (annotated via manager)
        all_staged_qs = import_job_model.stagedtransactionmodel_set.select_related(
            'parent',
            'account_model',
            'unit_model',
            'vendor_model',
            'customer_model',
            'transaction_model',
            'matched_transaction_model',
            'import_job',
            'import_job__ledger_model',
        )

        pending_qs = all_staged_qs.is_pending()
        imported_qs = all_staged_qs.is_imported()
        ready_qs = all_staged_qs.is_ready_to_import()

        total_count = all_staged_qs.count()
        imported_count = imported_qs.count()
        pending_count = pending_qs.count()
        ready_count = ready_qs.count()

        progress_pct = 0
        if total_count:
            progress_pct = round((imported_count / total_count) * 100)

        context.update(
            {
                'page_title': self.PAGE_TITLE,
                'header_title': self.PAGE_TITLE,
                'header_subtitle': import_job_model.description,
                'header_subtitle_icon': 'solar:import-bold',
                'staged_all_qs': all_staged_qs,
                'staged_pending_qs': pending_qs,
                'staged_imported_qs': imported_qs,
                'staged_ready_qs': ready_qs,
                'progress': {
                    'total': total_count,
                    'imported': imported_count,
                    'pending': pending_count,
                    'ready': ready_count,
                    'pct': progress_pct,
                },
            }
        )
        return context


class ImportJobModelUpdateView(ImportJobModelViewBaseView, UpdateView):
    template_name = 'django_ledger/data_import/import_job_update.html'
    context_object_name = 'import_job_model'
    pk_url_kwarg = 'job_pk'
    form_class = ImportJobModelUpdateForm

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        import_job_model: ImportJobModel = self.object
        ctx['page_title'] = 'Import Job Update'
        ctx['header_title'] = 'Import Job Update'
        ctx['header_subtitle'] = import_job_model.description
        ctx['header_subtitle_icon'] = 'solar:import-bold'
        return ctx

    def get_success_url(self):
        import_job_model: ImportJobModel = self.object
        return import_job_model.get_list_url()

    def form_valid(self, form):
        messages.add_message(
            self.request,
            level=messages.SUCCESS,
            message=_(f'Successfully updated Import Job {self.object.description}'),
            extra_tags='is-success',
        )
        return super().form_valid(form=form)


class ImportJobModelDeleteView(ImportJobModelViewBaseView, DeleteView):
    template_name = 'django_ledger/data_import/import_job_delete.html'
    context_object_name = 'import_job_model'
    pk_url_kwarg = 'job_pk'

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx['page_title'] = 'Delete Import Job'
        ctx['header_title'] = 'Delete Import Job'
        ctx['header_subtitle'] = self.object.description
        ctx['header_subtitle_icon'] = 'solar:import-bold'
        return ctx

    def get_success_url(self):
        import_job_model: ImportJobModel = self.object
        return import_job_model.get_list_url()


class ImportJobModelResetView(ImportJobModelViewBaseView, DetailView):
    pk_url_kwarg = 'job_pk'
    http_method_names = ['post']

    def post(self, request, **kwargs):
        import_job_model: ImportJobModel = self.get_object()
        staged_txs_qs = import_job_model.stagedtransactionmodel_set.all()

        with transaction.atomic():
            # First pass: clear matches and undo imports
            for staged_tx in staged_txs_qs:
                # Clear matched links (no effect if none or not allowed)
                try:
                    staged_tx.unmatch(raise_exception=False)
                except Exception:
                    # Ignore any unexpected issues while unmatching during a bulk reset
                    pass

                # Undo created imports / receipts if any
                staged_tx.undo_import(raise_exception=False)

            # Second pass: delete children to collapse splits
            for staged_tx in list(staged_txs_qs):
                if staged_tx.is_children():
                    staged_tx.delete()

        messages.add_message(
            request,
            messages.SUCCESS,
            _('Import job reset. All matches cleared and imports undone.'),
            extra_tags='is-success',
        )

        return redirect(
            to=import_job_model.get_data_import_url(),
            permanent=False,
        )


class StagedTransactionUpdateView(ImportJobModelViewBaseView, DetailView):
    template_name = 'django_ledger/data_import/staged_tx_update.html'
    PAGE_TITLE = _('Import Job Staged Txs')
    context_object_name = 'staged_tx'
    pk_url_kwarg = 'staged_tx_pk'
    import_transactions = False
    form_class = StagedTransactionModelFormSet
    http_method_names = ['get', 'post']
    extra_context = {
        'show_details': False,
    }

    def get_queryset(self):
        if self.queryset is None:
            self.queryset = (
                StagedTransactionModel.objects.for_entity(entity_model=self.AUTHORIZED_ENTITY_MODEL)
                .for_import_job(
                    import_job_model=self.kwargs['job_pk'],
                )
                .select_related(
                    'import_job',
                    'import_job__ledger_model',
                    'import_job__ledger_model__entity',
                )
                .is_pending()
                .is_parent()
            )
        return self.queryset

    def get_context_data(self, txs_formset: Optional[StagedTransactionModelFormSet] = None, **kwargs):
        context = super().get_context_data(**kwargs)

        staged_tx_model: StagedTransactionModel = self.object
        import_job_model: ImportJobModel = staged_tx_model.import_job

        context['page_title'] = import_job_model.description
        context['header_title'] = self.PAGE_TITLE
        context['header_subtitle'] = import_job_model.description
        context['header_subtitle_icon'] = 'tabler:table-import'

        staged_txs_formset = (
            StagedTransactionModelFormSet(
                entity_model=self.get_authorized_entity_instance(),
                import_job_model=import_job_model,
                staged_tx_pk=staged_tx_model,
            )
            if not txs_formset
            else txs_formset
        )

        context['formset'] = staged_txs_formset
        context['import_job_model'] = import_job_model

        return context

    def post(self, request, **kwargs):
        self.object = self.get_object()
        staged_tx: StagedTransactionModel = self.object
        import_job_model: ImportJobModel = staged_tx.import_job

        txs_formset = StagedTransactionModelFormSet(
            entity_model=self.AUTHORIZED_ENTITY_MODEL,
            import_job_model=import_job_model,
            staged_tx_pk=self.kwargs['staged_tx_pk'],
            data=request.POST,
        )
        is_import = False

        if txs_formset.has_changed():
            for tx_form in txs_formset:
                if any(
                    [
                        'account_model' in tx_form.changed_data,
                        'tx_split' in tx_form.changed_data,
                    ]
                ):
                    staged_transaction_model: StagedTransactionModel = tx_form.instance
                    staged_transaction_model.activity = None

            if txs_formset.is_valid():
                txs_formset.save()
                for tx_form in txs_formset:
                    if tx_form.has_changed():
                        staged_transaction_model: StagedTransactionModel = tx_form.instance

                        is_split = tx_form.cleaned_data['tx_split'] is True
                        is_import = tx_form.cleaned_data['tx_import'] is True
                        is_bundled = tx_form.cleaned_data['bundle_split'] is True

                        if is_split:
                            staged_transaction_model.add_split()
                        elif is_import:
                            selected_match = tx_form.cleaned_data.get('match_tx_model')
                            if selected_match is not None:
                                # Link to existing posted transaction via matched_transaction_model without creating a new JE
                                staged_transaction_model.matched_transaction_model = selected_match
                                staged_transaction_model.save(update_fields=['matched_transaction_model', 'updated'])
                            else:
                                if staged_transaction_model.can_migrate_receipt():
                                    staged_transaction_model.migrate_receipt(
                                        receipt_date=staged_transaction_model.date_posted,
                                        split_amount=not is_bundled,
                                    )
                                else:
                                    staged_transaction_model.migrate_transactions(split_txs=not is_bundled)

                # After processing, auto-collapse any parent that now has exactly one child
                parents_qs = (
                    import_job_model.stagedtransactionmodel_set.all()
                    .filter(parent__isnull=True)
                    .prefetch_related('split_transaction_set')
                )

                for parent in parents_qs:
                    children = list(parent.split_transaction_set.all())
                    if len(children) == 1:
                        child = children[0]
                        # Copy mapping fields from the remaining child back to the parent
                        parent.account_model = child.account_model
                        parent.unit_model = child.unit_model
                        parent.receipt_type = child.receipt_type
                        parent.vendor_model = child.vendor_model
                        parent.customer_model = child.customer_model
                        parent.bundle_split = True  # single transactions are treated as bundled
                        parent.save(
                            update_fields=[
                                'account_model',
                                'unit_model',
                                'receipt_type',
                                'vendor_model',
                                'customer_model',
                                'bundle_split',
                                'updated',
                            ]
                        )
                        # Remove the lone child so the parent becomes single again
                        child.delete()
            else:
                # formset not valid.....
                context = self.get_context_data(txs_formset=txs_formset, **kwargs)
                return self.render_to_response(context)

        messages.add_message(
            request,
            messages.SUCCESS,
            'Successfully saved transactions.',
            extra_tags='is-success',
        )

        if staged_tx.is_imported():
            return HttpResponseRedirect(redirect_to=import_job_model.get_detail_url())

        context = self.get_context_data(**kwargs)
        return self.render_to_response(context=context)


# Actions....
class StagedTransactionUndoView(ImportJobModelViewBaseView, View):
    http_method_names = ['post']

    def post(self, request, entity_slug, job_pk, staged_tx_pk, *args, **kwargs):
        import_job_model: ImportJobModel = get_object_or_404(self.get_queryset(), uuid__exact=job_pk)
        staged_txs_qs = import_job_model.stagedtransactionmodel_set.all()
        staged_tx = get_object_or_404(staged_txs_qs, uuid__exact=staged_tx_pk)
        try:
            staged_tx.undo_import()
            messages.add_message(
                request,
                messages.SUCCESS,
                _('Successfully undone import for transaction %(name)s.')
                % {'name': staged_tx.name or staged_tx.fit_id},
                extra_tags='is-success',
            )
        except StagedTransactionModelValidationError as e:
            messages.add_message(
                request,
                messages.ERROR,
                e.message,
                extra_tags='is-danger',
            )
        return redirect(
            to=import_job_model.get_detail_url(),
        )


class StagedTransactionUnmatchView(ImportJobModelViewBaseView, View):
    http_method_names = ['post']

    def post(self, request, entity_slug, job_pk, staged_tx_pk, *args, **kwargs):
        import_job_model: ImportJobModel = get_object_or_404(self.get_queryset(), uuid__exact=job_pk)
        staged_txs_qs = import_job_model.stagedtransactionmodel_set.all()

        staged_tx = get_object_or_404(staged_txs_qs, uuid__exact=staged_tx_pk)
        try:
            staged_tx.unmatch(commit=True)
            messages.add_message(
                request,
                messages.SUCCESS,
                _('Successfully cleared match for transaction %(name)s.')
                % {'name': staged_tx.name or staged_tx.fit_id},
                extra_tags='is-success',
            )
        except StagedTransactionModelValidationError as e:
            messages.add_message(
                request,
                messages.ERROR,
                e.message,
                extra_tags='is-danger',
            )
        return redirect(
            reverse(
                'django_ledger:import-job-detail',
                kwargs={'entity_slug': import_job_model.entity_slug, 'job_pk': import_job_model.uuid},
            )
        )


# Retired Views....



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
    * Miguel Sanda <msanda@arrobalytics.com>
"""

from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import CreateView, DetailView, ListView, UpdateView

from django_ledger.forms.vendor import VendorModelForm
from django_ledger.models.bill import BillModel
from django_ledger.models.entity import EntityModel
from django_ledger.models.receipt import ReceiptModel
from django_ledger.models.vendor import VendorModel
from django_ledger.views.mixins import DjangoLedgerSecurityMixIn


class VendorModelModelBaseView(DjangoLedgerSecurityMixIn):
    queryset = None

    def get_queryset(self):
        if self.queryset is None:
            self.queryset = VendorModel.objects.for_entity(
                entity_model=self.kwargs['entity_slug']
            ).order_by('-updated')
        return super().get_queryset()


class VendorModelListView(VendorModelModelBaseView, ListView):
    template_name = 'django_ledger/vendor/vendor_list.html'
    context_object_name = 'vendor_list'
    PAGE_TITLE = _('Vendor List')
    extra_context = {
        'page_title': PAGE_TITLE,
        'header_title': PAGE_TITLE,
        'header_subtitle_icon': 'bi:person-lines-fill',
    }

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        entity_model: EntityModel = self.get_authorized_entity_instance()
        context['header_subtitle'] = entity_model.name
        return context


class VendorModelCreateView(VendorModelModelBaseView, CreateView):
    template_name = 'django_ledger/vendor/vendor_create.html'
    PAGE_TITLE = _('Create New Vendor')
    form_class = VendorModelForm
    context_object_name = 'vendor'
    extra_context = {
        'page_title': PAGE_TITLE,
        'header_title': PAGE_TITLE,
        'header_subtitle_icon': 'bi:person-lines-fill',
    }

    def get_success_url(self):
        return reverse(
            'django_ledger:vendor-list',
            kwargs={'entity_slug': self.kwargs['entity_slug']},
        )

    def form_valid(self, form):
        vendor_model: VendorModel = form.save(commit=False)
        entity_model_qs = EntityModel.objects.for_user(user_model=self.request.user)
        entity_model = get_object_or_404(
            klass=entity_model_qs, slug__exact=self.kwargs['entity_slug']
        )
        vendor_model.entity_model = entity_model
        return super().form_valid(form)


class VendorModelUpdateView(VendorModelModelBaseView, UpdateView):
    template_name = 'django_ledger/vendor/vendor_update.html'
    PAGE_TITLE = _('Vendor Update')
    context_object_name = 'vendor'
    form_class = VendorModelForm

    slug_url_kwarg = 'vendor_pk'
    slug_field = 'uuid'

    def get_context_data(self, **kwargs):
        context = super(VendorModelUpdateView, self).get_context_data(**kwargs)
        vendor_model: VendorModel = self.object
        context['page_title'] = self.PAGE_TITLE
        context['header_title'] = self.PAGE_TITLE
        context['header_subtitle'] = vendor_model.vendor_number
        context['header_subtitle_icon'] = 'bi:person-lines-fill'
        return context

    def get_success_url(self):
        return reverse(
            'django_ledger:vendor-list',
            kwargs={'entity_slug': self.kwargs['entity_slug']},
        )


class VendorModelDetailView(VendorModelModelBaseView, DetailView):
    template_name = 'django_ledger/vendor/vendor_detail.html'
    context_object_name = 'vendor'
    PAGE_TITLE = _('Vendor Details')
    slug_url_kwarg = 'vendor_pk'
    slug_field = 'uuid'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        vendor_model: VendorModel = self.object
        receipts_qs = (
            ReceiptModel.objects.for_entity(entity_model=self.AUTHORIZED_ENTITY_MODEL)
            .for_vendor(vendor_model=vendor_model)
            .order_by('-receipt_date')
        )

        bills_qs = (
            BillModel.objects.for_entity(entity_model=self.AUTHORIZED_ENTITY_MODEL)
            .filter(vendor=vendor_model)
            .order_by('-updated')
        )

        context.update(
            {
                'page_title': self.PAGE_TITLE,
                'header_title': self.PAGE_TITLE,
                'header_subtitle': f'{vendor_model.vendor_name} · {vendor_model.vendor_number}',
                'header_subtitle_icon': 'bi:person-lines-fill',
                'receipts': receipts_qs,
                'bills': bills_qs,
            }
        )
        return context
